/*******************************************************************************
*   File:	astropci.c
*   Author:	Scott Streit
*   Abstract:	Solaris device driver for the SDSU PCI Interface Board.
*
*
*   Revision History:     Date      Who   Version    Description
*   --------------------------------------------------------------------------
*                       11/12/99    sds     1.0      Initial
*			12/09/99    sds     1.1      Made 64-bit Solaris 7 compatible
*						     and cleanup of unused variables.
*			01/04/2000  sds	    1.1      Made use of 64-bit offset in uio.
*			02/09/2000  sds     1.2      Added reply buffer ready/not_ready
*						     status checking to open() and close().
*						     Added polling in astropci_check_dsp_buffer()
*						     to prevent system hangup. Added check for
*						     abort readout in read().
*			02/17/2000  sds	    1.2	     Removed debug print statements form read().
*			02/21/2000  sds	    1.2	     Changed the DSP FIFO Empty bit check
*						     from bit1 to bit0.
*			02/24/2000  sds     1.2      Removed ALL reply buffer ready/not_ready
*						     status checking. Replaced get_reply()
*						     function with a new one that uses a signal
*						     and variable check.
*			04/20/2000  sds     1.3      Added the ioctl command ASTROPCI_FLUSH_REPLY_BUFFER
*						     to allow the user application to reset the
*						     reply buffer after a hardware reset (RST/SYR).
*						     Changed the DSP FIFO Empty bit check from bit1
*						     to bit0.
*			04/25/2000  sds     1.3      Removed all but one check of the status buffer in
*						     the ISR (astropci_intr). Also added checks to
*						     handle re-entry into the ISR by using mutex_tryenter(),
*						     stray interrupts are caught, and stray reply interrupts
*						     that contain no data are caught.
*			06/21/2000  sds     1.3      Removed the useless ioctl command ASTROPCI_GET_CMDR. The
*						     manual command register should only be set, not read.
*			06/27/2000  sds     1.4      Added ioctl binning commands ASTROPCI_SET_BINNING_ROWS
*						     and ASTROPCI_SET_BINNING_COLS.
*			06/28/2000  sds     1.4      Added ioctl binning commands ASTROPCI_SET_BINNING_ROWS
*						     and ASTROPCI_SET_BINNING_COLS. Added the read replies
*						     after setting the kernel buffer addresses.
*			07/13/2000  sds     1.4      Added byte swapping of the data words before transfer
*						     to user space.
*			07/20/2000  sds     1.4      Added a PCI DSP program counter reset function call to
*						     the open() function to be sure the PCI board is ready.
*			07/25/2000  sds     1.4      Removed the get_reply() calls after setting the image
*						     buffers.
*
*   Development notes:
*   -----------------
*   This driver has been crudely tested on Solaris 2.6 and it appears to
*   function properly, however, the driver will not be supported on that OS
*   version.  This driver is only supported for Solaris 7.
*
*   No standard include files outside of the <sys/...> directory can be
*   included since this code runs in kernel space.  Only functions and
*   structures in Section 9 of the manual are usable here.  (Yes, that
*   means that <string.h>, <stdio.h>, and friends and there functions
*   cannot be used).
*
*   The header files <sys/ddi.h> and <sys/sunddi.h> must be at the end
*   of the include file list.  These headers undefine several macros
*   that are re-implemented as functions.
*******************************************************************************/
#include <sys/conf.h>
#include <sys/devops.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/kmem.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>
#include "astropci_impl.h"
#include "astropci_io.h"
#include <sys/uio.h>
#include <sys/map.h>
#include <sys/debug.h>
#include <sys/cmn_err.h>
#include <sys/ddi.h>		/* these two must be last */
#include <sys/sunddi.h>

/*******************************************************************************
	Prototypes for main entry points pointed to in the cb_ops structure
*******************************************************************************/
static int astropci_open(dev_t *devp, int flag, int otyp, cred_t *credp);
static int astropci_close(dev_t dev, int flag, int otyp, cred_t *credp);
static int astropci_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int astropci_ioctl(dev_t dev, int cmd, intptr_t arg, int mode, cred_t *credp, int *rvalp);

/******************************************************************************* 
	Prototypes for functions pointed to by the dev_ops structure
*******************************************************************************/
static int astropci_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result);
static int astropci_identify(dev_info_t *dip);
static int astropci_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int astropci_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

/*******************************************************************************
	Prototypes for user defined functions
*******************************************************************************/
static u_int astropci_intr(caddr_t instance);
static int astropci_read_reply(struct astropci_state *asp);
static int astropci_check_dsp_buffer(struct astropci_state *asp);
static int astropci_set_buffer_addresses(struct astropci_state *asp);
static int astropci_set_reply_address(struct astropci_state *asp);
static void astropci_set_reply_buffer_status(struct astropci_state *asp, int status);

/*******************************************************************************
	statep - Anchor point for soft state information.
*******************************************************************************/
static void *statep;			/* opaque */

/*******************************************************************************
	cb_ops - character/block entry points structure. Contains all entry
		 points for  drivers  that  support both  character  and
		 block  entry  points. "nodev" specifies that the entry point
		 is not in use.
*******************************************************************************/
static struct cb_ops astropci_cb_ops = {
	astropci_open,			/* XX_open()     */
 	astropci_close,			/* XX_close()    */
 	nodev,				/* XX_strategy() */
  	nodev,				/* XX_print()    */
  	nodev,				/* XX_dump()     */
  	astropci_read,			/* XX_read ()    */
  	nodev,				/* XX_write()    */
  	astropci_ioctl,			/* XX_ioctl()    */
  	nodev,				/* XX_devmap()   */
  	nodev,				/* XX_mmap()     */
  	nodev,				/* XX_segmap()   */
  	nochpoll,			/* XX_chpoll()   */
  	ddi_prop_op,			/* XX_prop_op()  */
  	NULL,
  	(D_NEW | D_MP | D_64BIT)	/* Flags */
};

/*******************************************************************************
	dev_ops - device operations structure. Contains driver common fields
		  and  pointers  to  the bus_ops and/or cb_ops(9S). "nulldev"
		  specifie no value.
*******************************************************************************/
static struct dev_ops astropci_ops = {
	DEVO_REV,			/* Driver build version      */
  	0,				/* Driver reference count    */
  	astropci_getinfo,		/* Get driver information    */
  	astropci_identify,		/* Identify associated device*/
  	nulldev,			/* Probe device	             */
  	astropci_attach,		/* Attach driver to dev_info */
  	astropci_detach,		/* Unload/detach driver      */
  	nodev,				/* Reset device              */
  	&astropci_cb_ops,		/* Pointer to cb_ops struct  */
  	NULL				/* Pointer to bus_ops struct */
};

/*******************************************************************************
	modldrv - linkage structure for loadable drivers. The modldrv structure
		  is used by device  drivers  to  export driver specific
		  information to the kernel.
*******************************************************************************/
extern struct mod_ops mod_driverops;
static struct modldrv modldrv = {
	&mod_driverops,
	"astropci driver v1.0",
	&astropci_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
  	&modldrv,
  	NULL
};

/*******************************************************************************
 FUNCTION: _INIT
 
 PURPOSE:  Entry point. "_init(9e)" initializes a loadable module.
 
 RETURNS:  Returns an appropriate error number if there is an error, else it
 	   returns the return value from mod_install(9F).
 
 NOTES:    It is called before any other routine and returns the evaluation of
           mod_install(9f).  Any other work should be done prior to calling
           mod_install(9f) and must be undone if mod_install(9f) fails. Called
           by the kernel during module loading and not available to user code.
*******************************************************************************/
int _init(void)
{
	int err = ddi_soft_state_init (&statep, sizeof (struct astropci_state), 1);

	/* Check that the soft state can be initialized */
   	if (err) {
   		cmn_err(CE_WARN, "State info cannot be initialized.");
   		return err;
   	}
    
   	err = mod_install (&modlinkage);

  	 if (err) {
   		ddi_soft_state_fini (&statep);
   		cmn_err(CE_WARN, "Driver cannot be initialized.");
   	}

   	return err;
}

/*******************************************************************************
 FUNCTION: _INFO
 
 PURPOSE:  Entry point. "_info(9e)" returns information about a loadable module.
 
 RETURNS:  Returns the return value from mod_info(9F).
 
 NOTES:    Called by the kernel during module loading and not available to user
 	   code.
*******************************************************************************/
int _info (struct modinfo *modinfop)
{
	int error;

	error = mod_info (&modlinkage, modinfop);

   	return error;
}

/*******************************************************************************
 FUNCTION: _FINI
 
 PURPOSE:  Entry point. "_fini(9e)" prepares the module for unloading.
 
 RETURNS:  Returns the return value from mod_info(9F).
 
 NOTES:    It should deallocate any resources allocated by _init(9e). Called
           by the kernel during module loading and not available to user code.
*******************************************************************************/
int _fini (void)
{
	int err = mod_remove (&modlinkage);

   	if (!err)
   		ddi_soft_state_fini (&statep);
   	else
   		cmn_err(CE_WARN, "Driver cannot be unloaded.");
    
   	return err;
}


/*******************************************************************************
 FUNCTION: ASTROPCI_GETINFO
 
 PURPOSE:  Entry point. Get device driver information.
 
 RETURNS:  Returns DDI_SUCCESS on success, DDI_FAILURE on failure.
 
 NOTES:    There is an implicit mapping here between the instance number
           and the minor number such that instance = minor number.
*******************************************************************************/
static int astropci_getinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	struct astropci_state *xsp;
  	dev_t dev = (dev_t) arg;
  	int instance;

  	switch (infocmd)
  	{
  		case DDI_INFO_DEVT2INSTANCE:
        		*result = (void *)(intptr_t)getminor(dev);
        		return DDI_SUCCESS;
    		
    		case DDI_INFO_DEVT2DEVINFO:
			instance = getminor(dev);
			xsp = ddi_get_soft_state(statep, instance);

        		if (xsp == NULL)
	  			return DDI_FAILURE;

			*result = (void *) xsp->dip;
			return DDI_SUCCESS;
    		
    		default:
      			return DDI_FAILURE;
  	}
}

/*******************************************************************************
 FUNCTION: ASTROPCI_IDENTIFY
 
 PURPOSE:  Entry point. Determine if a driver is associated with a device.
 
 RETURNS:  The return value from xxidentify() is ignored.
 
 NOTES:    
*******************************************************************************/
static int astropci_identify (dev_info_t *dip)
{
	int value;
 
  	if (strcmp (ddi_get_name (dip), prom_id_string_one) == 0) {
    		cmn_err(CE_NOTE, "Device identified as %s", prom_id_string_one);
    		value = DDI_IDENTIFIED;
  	}
  	
  	else if (strcmp (ddi_get_name (dip), prom_id_string_two) == 0) {
    		cmn_err(CE_NOTE, "Device identified as %s", prom_id_string_two);
    		value = DDI_IDENTIFIED;
  	}
  	
  	else {
    		cmn_err (CE_WARN, "Device NOT identified.");
    		value = DDI_NOT_IDENTIFIED;
  	}

  	return value;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_ATTACH
 
 PURPOSE:  Entry point. Attach a device to the system, or resume it.
 
 RETURNS:  Returns DDI_SUCCESS on success, DDI_FAILURE on failure.
 
 NOTES:    For a character device (memory-mapped), attach(9e) needs to:
           1. Allocate a state structure and initialize it
           2. Map the device's registers
           3. Add the interrupt handler
           4. Initialize mutexes and condition variables
           5. Create the device's minor node
*******************************************************************************/
static int astropci_attach (dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	struct astropci_state *asp = NULL;
	int instance;
  	int success;
		
	instance = ddi_get_instance(dip);
	success = DDI_SUCCESS;

  	switch (cmd)
  	{
		case DDI_ATTACH:
        		/* Quit if device is slave-only -- can't do DMA then */
			if (ddi_slaveonly (dip) == DDI_SUCCESS) {
           			cmn_err(CE_WARN, "astropci: Cannot attach device that is slave-only.");
	    			return DDI_FAILURE;
	  		}

			/* Quit if interrupt level is too high */
			if (ddi_intr_hilevel (dip, 0) != 0) {
	    			cmn_err (CE_WARN, "astropci: High-level interrupts are not supported");
	    			return DDI_FAILURE;
	  		}
	  		
			/* Allocate the soft state structure */
			if (ddi_soft_state_zalloc (statep, instance) != 0) {
            			cmn_err(CE_NOTE, "astropci: Cannot allocate the soft state structure.");
	    			return DDI_FAILURE;
	  		}
		
			/* Assign the soft state structure */
			asp = ddi_get_soft_state (statep, instance);
			asp->dip = dip;
			asp->busy = 0;
			ddi_set_driver_private(dip, (caddr_t)asp);

        		if (asp == NULL)
        			cmn_err(CE_NOTE, "astropci: Soft state structure allocation failed.");

			/* Map the device configuration registers */
			if (pci_config_setup(dip, &asp->config_access_handle) != DDI_SUCCESS) {
            			cmn_err(CE_WARN, "astropci: Cannot map the device configuration registers.");
	    			success = DDI_FAILURE;
	    			break;
	  		}

			/* Map the device memory (DSP/PCI control) registers */
			if (ddi_regs_map_setup ( dip, 1, (caddr_t *)&asp->regp, 0, sizeof(struct astropci_regs), &access_attr, &asp->reg_access_handle) != DDI_SUCCESS) {
            			cmn_err(CE_WARN, "astropci: Cannot map the device memory (DSP/PCI control) registers.");
 	    			success = DDI_FAILURE;
	    			break;
	  		}

			/* Map the device memory (camera control) registers */
			if (ddi_regs_map_setup ( dip, 1, (caddr_t *)&asp->camregp, 0x80, sizeof(struct astropci_camregs), &access_attr, &asp->camreg_access_handle) != DDI_SUCCESS) {
            			cmn_err(CE_WARN, "astropci: Cannot map the device memory (camera control) registers.");
 	    			success = DDI_FAILURE;
	    			break;
	  		}
		
			/* Get an interrupt block cookie */
			ddi_get_iblock_cookie(dip, 0, &asp->iblock_cookie);

			/* Initialize the mutex locks and condition variables */
			mutex_init (&asp->mu, "astropci mutex", MUTEX_DRIVER, (void *)&asp->iblock_cookie);
			mutex_init (&asp->reply_mu, "astropci reply mutex", MUTEX_DRIVER, (void *)&asp->iblock_cookie);
			cv_init (&asp->cv, NULL, CV_DRIVER, NULL);
			cv_init (&asp->reply_cv, NULL, CV_DRIVER, NULL);
			cv_init (&asp->bufcv, NULL, CV_DRIVER, NULL);
			cv_init (&asp->reply_bufcv, NULL, CV_DRIVER, NULL);
			cv_init (&asp->intrcv, NULL, CV_DRIVER, NULL);

			/* Register the interrupt handler */
        		if (ddi_add_intr (dip, 0, &asp->iblock_cookie, (ddi_idevice_cookie_t *)NULL, astropci_intr, (caddr_t)(intptr_t)instance) != DDI_SUCCESS) {
            			cmn_err(CE_WARN, "astropci: Cannot register interrupt handler.");
            			success = DDI_FAILURE;
            			break;
			}

	       		/* Create minor node */
			if (ddi_create_minor_node (dip, "astropci0", S_IFCHR, instance,"ddi_sdsu_astropci", 0) != DDI_SUCCESS) {
            			cmn_err(CE_WARN, "astropci: Cannot create minor node 0.");
            			success = DDI_FAILURE;
            			break;
	  		}

			/* Create another minor node (for a dual read-out board) */
			if (ddi_create_minor_node (dip, "astropci1", S_IFCHR, instance+1,"ddi_sdsu_astropci", 0) != DDI_SUCCESS) {
            			cmn_err(CE_WARN, "astropci: Cannot create minor node 1.");
	    			success = DDI_FAILURE;
	    			break;
	  		}

			/* Complete initializing soft state structure */
			asp->dip = dip;
			asp->open = 0;

			break;

		default:
			cmn_err(CE_WARN, "astropci: Cannot attach device driver.");
			return DDI_FAILURE;
	}

	/* Deallocate any resources allocated above */
	if (success == DDI_FAILURE) {
		/* Remove all minor nodes */
        	ddi_remove_minor_node (dip, NULL);
        	
 		/* Free all registered interrupts */
        	ddi_remove_intr (dip, 0, asp->iblock_cookie);

		/* Free the configuration registers */
		if (asp->config_access_handle != NULL)
			pci_config_teardown(&asp->config_access_handle);
		
		/* Free the DSP memory mapped registers */
		if (asp->regp != NULL)
			ddi_regs_map_free(&asp->reg_access_handle);

		/* Destroy all mutexes and condition variables */
		cv_destroy (&asp->cv);
        	cv_destroy (&asp->reply_cv);
        	cv_destroy (&asp->bufcv);
		cv_destroy (&asp->reply_bufcv);
       		mutex_destroy (&asp->mu);
        	mutex_destroy (&asp->reply_mu);

     		/* Free the soft state */
     		if (asp != NULL)
     			ddi_soft_state_free (statep, instance);
  	}
 
  	else
		cmn_err(CE_NOTE, "astropci: Device driver successfully attached.");

	return success;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_DETACH
 
 PURPOSE:  Entry point. Detach a device.
 
 RETURNS:  Returns DDI_SUCCESS on success, DDI_FAILURE on failure.
 
 NOTES:    Essentially, detach(9e) should undo all the resource allocations
           that attach(9e) successfully completes.
*******************************************************************************/
static int astropci_detach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int instance = ddi_get_instance (dip);
  	struct astropci_state *asp = ddi_get_soft_state (statep, instance);

  	switch (cmd)
  	{
		case DDI_DETACH:
        		/* Remove all minor nodes */
        		ddi_remove_minor_node (dip, NULL);
        	
 			/* Free all registered interrupts */
        		ddi_remove_intr (dip, 0, asp->iblock_cookie);

			/* Free the configuration and DSP memory mapped registers */
			pci_config_teardown(&asp->config_access_handle);
			if (asp->regp != NULL)
				ddi_regs_map_free(&asp->reg_access_handle);

        		/* Destroy all mutexes and condition variables */
        		cv_destroy (&asp->cv);
        		cv_destroy (&asp->reply_cv);
        		cv_destroy (&asp->bufcv);
			cv_destroy (&asp->reply_bufcv);
        		mutex_destroy (&asp->mu);
        		mutex_destroy (&asp->reply_mu);

       			/* Free the soft state */
       			if (asp != NULL)
        			ddi_soft_state_free (statep, instance);
			break;

		default:
			return DDI_FAILURE;
  	}
  	
	cmn_err(CE_NOTE, "astropci: Device driver successfully dettached.");
  	return DDI_SUCCESS;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_OPEN
 
 PURPOSE:  Entry point. Open a device for access. The image buffers are
 	   allocated here.
 
 RETURNS:  Returns 0 for success, or the appropriate error number.
 
 NOTES:    
*******************************************************************************/
static int astropci_open (dev_t *devp, int flag, int otyp, cred_t *credp)
{
	int x = 0;
  	int value = 0;
    	int reply_value = 0;
	uint32_t hctr_value = 0;
	uint32_t reply_status = 0;

  	int instance = getminor (*devp);
  	struct astropci_state *asp = ddi_get_soft_state (statep, instance);

  	if (asp == NULL)
    		return ENXIO;

  	if (otyp != OTYP_CHR)
    		return EINVAL;

  	mutex_enter (&asp->mu);

  	/* Honor exclusive lock flag */  
  	if (flag == FEXCL && asp->open > 0)
  		return EAGAIN;

  	asp->open++;

	/* Create image buffers */
	value = astropci_set_buffer_addresses(asp);

	/* Clear the reply */
	asp->reply = REPLY_BUFFER_EMPTY;

	/* Set HCTR bits 7, 11, and 12 to 0 for reply buffer setup */
	hctr_value = ddi_get32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr);
	ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr,(uint32_t)(hctr_value & 0x77F));

	/* Flush the reply buffer FIFO on the PCI DSP. 3 is the number of words the FIFO can hold */
   	for (x=0; x<3; x++) {
	  	reply_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hstr);
   		if (reply_status & REPLY_DATA_EXISTS)
  			reply_value = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->reply_buffer);
  		else
  			break;
  	}

	mutex_exit (&asp->mu);

	return value;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_CLOSE
 
 PURPOSE:  Entry point. Close a device from access.
 
 RETURNS:  Returns 0 for success, or the appropriate error number.
 
 NOTES:    
*******************************************************************************/
static int astropci_close (dev_t dev, int flag, int otyp, cred_t *credp)
{
	struct image_buf *ip;
  	struct reply_buf *rp;
  	int j = 0;
  	int instance = getminor (dev);
  	struct astropci_state *asp = ddi_get_soft_state (statep, instance);

  	if (asp == NULL)		/* sanity checks */
    		return ENXIO;

  	if (otyp != OTYP_CHR)
    		return EINVAL;

  	mutex_enter (&asp->mu);

  	asp->open = 0;

  	for (j=0; j<2; j++) {
  		if (asp->image_buf[j].buf_handle != (caddr_t) -1) {
     			ip = &asp->image_buf[j];
     			if ( ddi_dma_unbind_handle(ip->buf_handle) != DDI_SUCCESS )
         			cmn_err(CE_WARN,"Unable to unbind image buffer %d handle.", j);
     			else {
         			ddi_dma_mem_free(&ip->data_access_handle);
         			ddi_dma_free_handle(&ip->buf_handle);
     			}
  		}
  	}
	
  	mutex_exit (&asp->mu);
  
  	return 0;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_READ
 
 PURPOSE:  Entry point. Read data from a device.
 
 RETURNS:  Returns 0 for success, or the appropriate error number. ECHILD
 	   is returned if a buffer is in an unexpected state. EIO is returned
 	   if a timeout occurs while waiting for image data.
 
 NOTES:    
*******************************************************************************/
static int astropci_read (dev_t dev, struct uio *uiop, cred_t *credp)
{
	struct image_buf *ip;
 	int value = 0;
  	long sleeping;
  	uint32_t transfer_done;
  	clock_t time1;
  	clock_t time2;
  	uint32_t transfer_status;
	uint32_t temp1 = 0;
	uint32_t temp2 = 0;
	uint32_t temp3 = 0;
	uint32_t temp4 = 0;
 	long address;
 	int j;
 	int i;

  	/* Get current soft state */
  	int instance = getminor (dev);
  	struct astropci_state *asp = ddi_get_soft_state (statep, instance);
  	if (asp == NULL)
     		return ENXIO;

  	/* Allow only one thread at a time */
  	mutex_enter (&asp->mu);
  	while (asp->busy) {
		cv_wait (&asp->cv, &asp->mu);
  	}
  	asp->busy = 1;
  	mutex_exit (&asp->mu);

  	/* Get the size of the data to be transferred from the user */
  	asp->byteremain = uiop->uio_resid;

  	/* Initialize the image buffer states */
  	for (j=0; j<2; j++)
		asp->image_buf[j].state = image_buf_waiting;

  	/* Initialize the buffer states */
  	asp->buf_ready = 0;
  	asp->buf_busy = 0;
  	asp->buf_done = 0;
  	asp->dma_done = 0;
  	transfer_done = 0;
  	asp->num_of_bufs_in_use = 0;
  	asp->progress_buffer = 0;  
  	asp->dma_counter = 0;
  	asp->abort = 0;
  	
  	/* Set the image transfer status to 1 */
  	transfer_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hctr);
  	ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr, (uint32_t)(transfer_status | 0x8));

  	mutex_enter(&asp->mu);
  	while (transfer_done == 0) {
		ip = &asp->image_buf[asp->buf_ready];
		
		/* Check that the buffer is in the correct state */
		if (ip->state != image_buf_waiting) {
  			value = ECHILD;
			goto EXITREAD;
		}
  			
  		/* Set the size of the data to be transferred */
  		if ((asp->byteremain) > (ulong_t)(ip->length))
  			ip->size = ip->length;
  		else
  			ip->size = asp->byteremain;

        	/* Set the number of bytes remaining to be transferred */
       		asp->byteremain = asp->byteremain - ip->size;

  		asp->num_of_bufs_in_use++;
  		ip->state = image_buf_busy;
  		asp->buf_busy = asp->buf_ready;
		asp->buf_ready++;
		if (asp->buf_ready > 1)
			asp->buf_ready = 0;

  		/* Wait for DMA process and interrupt routine to finish */
  		while (asp->dma_done != 1 && asp->dma_counter == 0) {
  			
  			/* Check for abort */
  			if (asp->abort) {
  				cmn_err (CE_WARN,"Readout Aborted (1).");
  				value = EINTR;
  				mutex_exit(&asp->mu);
  				goto EXITREAD;
  			}
  			
  			drv_getparm(LBOLT, (u_long *)&time1);
			time2 = time1 + drv_usectohz(dma_read_timeout_usec);
			sleeping = cv_timedwait_sig(&asp->bufcv,&asp->mu,time2);
			
			/* Check for DMA timeout and exit if a timeout has occurred */
			if (sleeping == -1) {
				cmn_err (CE_WARN,"astropci_read: DMA timeout!");
				value = EIO;
				mutex_exit(&asp->mu);
				goto EXITREAD;
			}	
		}

		/* Check for abort */
		if (asp->abort) {
			cmn_err (CE_WARN,"Readout Aborted (2).");
  			value = EINTR;
			mutex_exit(&asp->mu);
			goto EXITREAD;
		}

#if 0
		/* Swap the image data bytes within a word, do four bytes at a time. */
		for (i=0; i<(ip->length)-4; i+=4) {
			temp1 = ip->address[i];
			temp2 = ip->address[i+1];
			temp3 = ip->address[i+2];
			temp4 = ip->address[i+3];

			ip->address[i] = temp2;
			ip->address[i+1] = temp1;
			ip->address[i+2] = temp4;
			ip->address[i+3] = temp3;
		}
#endif
  				
		/* Transfer the image data to user space */
  		/*if (ddi_dma_sync(ip->buf_handle, 0, ip->size, DDI_DMA_SYNC_FORCPU) != DDI_SUCCESS)
			cmn_err (CE_WARN,"astropci_read: Unable to sync buffer");
		else {*/
				address = (intptr_t)ip->address;
		  		asp->progress_buffer++;

				for (i=0; i<256; i++) {
					
					/* Check for abort */
					if (asp->abort) {
						cmn_err (CE_WARN,"Readout Aborted (3).");
  						value = EINTR;
						mutex_exit(&asp->mu);
						goto EXITREAD;
					}

	       				if (uiomove((caddr_t)address, 512, UIO_READ, uiop) != 0)
       						cmn_err (CE_WARN,"astropci_read: Copy error to user space");
       					address = address + 512;

	       				/*if (uiomove((caddr_t)address, 131072, UIO_READ, uiop) != 0)
       						cmn_err (CE_WARN,"astropci_read: Copy error to user space");
       					address = address + 131072;*/
       				}
       		/*}*/

		/* Check for abort */
		if (asp->abort) {
			cmn_err (CE_WARN,"Readout Aborted (4).");
  			value = EINTR;
			mutex_exit(&asp->mu);
			goto EXITREAD;
		}

		/* Clear the image transfer status */
		transfer_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hctr);
		if (asp->buf_busy == 0) 
			ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr, (uint32_t)(transfer_status & IMAGE_BUFF1_CLEAR));
		else if (asp->buf_busy == 1) 
			ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr, (uint32_t)(transfer_status & IMAGE_BUFF2_CLEAR));

		/* Check for abort */
		if (asp->abort) {
			cmn_err (CE_WARN,"Readout Aborted (5).");
  			value = EINTR;
			mutex_exit(&asp->mu);
			goto EXITREAD;
		}

		/* Re-initialize variables */
		ip->state = image_buf_waiting;
		asp->num_of_bufs_in_use--;
		asp->dma_done = 0;
		asp->dma_counter--;
			
		/* Check for completion of image transfer */
		if (asp->byteremain == 0)
			transfer_done = 1;
  	}

	mutex_exit (&asp->mu);

EXITREAD:
  	asp->abort = 0;
  	asp->busy = 0;
  	asp->progress_buffer = 0;  
	cv_signal(&asp->cv);
  
  	return value;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_READ_REPLY
 
 PURPOSE:  Support function. Read command reply from DSP.
 
 RETURNS:  Returns 0 for success, or the appropriate error number.
 
 NOTES:    
*******************************************************************************/
static int astropci_read_reply (struct astropci_state *asp)
{
  	uint32_t ret_value = 0;
  	long sleeping = 0;
  	clock_t time1 = 0;
  	clock_t time2 = 0;
  
	if (asp->reply == REPLY_BUFFER_EMPTY) {
		/* Wait for a signal if no reply */
		drv_getparm(LBOLT, (u_long *)&time1);
		time2 = time1 + drv_usectohz(read_reply_timeout_usec);
		sleeping = cv_timedwait_sig(&asp->reply_bufcv, &asp->mu, time2);
	}
	
	ret_value = asp->reply;
	asp->reply = REPLY_BUFFER_EMPTY;
	
	/* Check for a timeout and exit if one has occurred */
	if (sleeping == -1) {
		cmn_err (CE_WARN,"astropci_read_reply: Reply timeout!");
		ret_value = TIMEOUT;
	}

  	return ret_value;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_IOCTL
 
 PURPOSE:  Entry point. Control a character device.
 
 RETURNS:  Returns 0 for success, or the appropriate error number.
 
 NOTES:    Locking should be used on a case by case basis, since it is not
	   always necessary. Calls to allocation routines that might sleep
	   should never be locked.

           ddi_copyin(9f) and ddi_copyout(9f) move data from/to user space
           to/from kernel space respectively. These must be used for this
           purpose.
*******************************************************************************/
static int astropci_ioctl (dev_t dev, int cmd, intptr_t arg, int mode, cred_t *credp, int *rvalp)
{
	int instance = getminor(dev);
  	struct astropci_state *asp = ddi_get_soft_state(statep, instance);
  	struct reply_buf *rp;
  	int retval = 0;

  	/* Sanity check */
  	if (asp == NULL)
  		return ENXIO;

  	mutex_tryenter (&asp->mu);

  	switch (cmd)
    	{
    		case ASTROPCI_GET_HCTR:
    		    {
			uint32_t hctr = ddi_get32(asp->reg_access_handle,(uint32_t *)&asp->regp->hctr);
        		if (ddi_copyout((caddr_t) &hctr, (caddr_t) arg, sizeof (hctr), mode) != 0)
	   			retval = EFAULT;
      		    }
      		    break;

    		case ASTROPCI_GET_HSTR:
    		    {
			uint32_t hstr = ddi_get32(asp->reg_access_handle,(uint32_t *)&asp->regp->hstr);
        		if (ddi_copyout((caddr_t) &hstr, (caddr_t) arg, sizeof (hstr), mode) != 0)
	   			retval = EFAULT;
	      	    }
	      	    break;
            
    		case ASTROPCI_GET_PROGRESS:
      		    {
       			uint32_t progress_buffer = asp->progress_buffer; /* ddi_get32(asp->reg_access_handle,(uint32_t *)&asp->progress_buffer);*/
       		
       			if (ddi_copyout((caddr_t)&progress_buffer, (caddr_t)arg, sizeof(progress_buffer), mode) != 0)
	   			retval = EFAULT;
      		    }
      		    break;
      
    		case ASTROPCI_SET_HCTR:
      		    {
			uint32_t hctr;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &hctr, sizeof(hctr), mode) != 0)
	   			retval = EFAULT;
			else
	   			ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr,(uint32_t)hctr);
		
			/* Wait 10 microseconds */
			drv_usecwait(10);
      		    }  		
     		    break;

    		case ASTROPCI_SET_HCVR:
      		    {
			uint32_t hcvr;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &hcvr, sizeof(hcvr), mode) != 0)
	   			retval = EFAULT;
			else
	   			ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr,(uint32_t)hcvr);
	
			/* Wait 10 microseconds */
			/* drv_usecwait(10); */
      		    }  		
      		    break;
      
    		case ASTROPCI_SET_CMDR:
      		    {
        		uint32_t manual_cmd;

        		if (ddi_copyin((caddr_t) arg, (caddr_t)&manual_cmd, sizeof(manual_cmd), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->manual_cmd, manual_cmd);
	   			else
	   				retval = EIO;
	   		}
      		    }
      		    break;

		case ASTROPCI_SET_ARG1:
      		    {
			uint32_t arg1;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &arg1, sizeof(arg1), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->arg1,(uint32_t)arg1);
	   			else
	   				retval = EIO;
			}
      		    }
      		    break;

    		case ASTROPCI_SET_ARG2:
      		    {
			uint32_t arg2;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &arg2, sizeof(arg2), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->arg2,(uint32_t)arg2);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;

    		case ASTROPCI_SET_ARG3:
      		    {
			uint32_t arg3;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &arg3, sizeof(arg3), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->arg3,(uint32_t)arg3);
	   			else
	   				retval = EIO;
        		}
      		    }
      		    break;

    		case ASTROPCI_SET_ARG4:
      		    {
			uint32_t arg4;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &arg4, sizeof(arg4), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->arg4,(uint32_t)arg4);
 	   			else
	   				retval = EIO;
	       		}
      		    }
      		    break;

    		case ASTROPCI_SET_ARG5:
      		    {
			uint32_t arg5;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &arg5, sizeof(arg5), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->arg5,(uint32_t)arg5);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;

    		case ASTROPCI_SET_EXPTIME:
      		    {
			uint32_t exp_time;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &exp_time, sizeof(exp_time), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->camreg_access_handle, (uint32_t *)&asp->camregp->exptime,(uint32_t)exp_time);
	   			else
	   				retval = EIO;
      			}
       		    }
      		    break;

    		case ASTROPCI_SET_NCOLS:
      		    {
			uint32_t columns;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &columns, sizeof(columns), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->camreg_access_handle, (uint32_t *)&asp->camregp->ncols,(uint32_t)columns);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;

    		case ASTROPCI_SET_NROWS:
      		    {
			uint32_t rows;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &rows, sizeof(rows), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->camreg_access_handle, (uint32_t *)&asp->camregp->nrows,(uint32_t)rows);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;

    		case ASTROPCI_SET_BINNING_COLS:
      		    {
			uint32_t bin_cols;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &bin_cols, sizeof(bin_cols), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->camreg_access_handle, (uint32_t *)&asp->camregp->ncols_bin,(uint32_t)bin_cols);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;

    		case ASTROPCI_SET_BINNING_ROWS:
      		    {
			uint32_t bin_rows;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &bin_rows, sizeof(bin_rows), mode) != 0)
	   			retval = EFAULT;
			else {
	   			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
	   				ddi_put32(asp->camreg_access_handle, (uint32_t *)&asp->camregp->nrows_bin,(uint32_t)bin_rows);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;
      		          
    		case ASTROPCI_SET_DESTINATION:
      		    {
			uint32_t board_destination;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &board_destination, sizeof(board_destination), mode) != 0)
	   			retval = EFAULT;
      			else {
      				if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
      					ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->board_dest, board_destination);
	   			else
	   				retval = EIO;
      			}
      		    }
      		    break;

    		case ASTROPCI_SET_IMAGE_BUFFERS:
      		    {
			/* Pass the DMA kernel buffer addresses to the DSP */
  			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
  				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->dma_addr1, (uint32_t)asp->image_buffer_address1);
   			else
   				retval = EIO;

  			if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
  				ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->dma_addr2, (uint32_t)asp->image_buffer_address2);
   			else
   				retval = EIO;
      
        		if (ddi_copyout((caddr_t) &asp->image_buffer_address2, (caddr_t) arg, sizeof (asp->image_buffer_address2), mode) != 0)
	   			retval = EFAULT;
      		    }
      		    break;

    		case ASTROPCI_SET_UTIL_OPTIONS:
      		    {
			uint32_t utility_options;

			if (ddi_copyin((caddr_t) arg, (caddr_t) &utility_options, sizeof(utility_options), mode) != 0)
	   			retval = EFAULT;
      			else {
      				if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
      					ddi_put32(asp->camreg_access_handle, (uint32_t *)&asp->camregp->util_options, utility_options);
	   			else
	   				retval = EIO;
      			}
       		    }
      		    break;


   	 	case ASTROPCI_GET_REPLY:
    		    {
   			int reply = 0;
   	
			reply = astropci_read_reply(asp);
   	
   			if (ddi_copyout((caddr_t) &reply, (caddr_t) arg, sizeof(reply), mode) != 0)
	   			retval = EFAULT;
     		    }
     		    break;

    		case ASTROPCI_ABORT_READ:
			cmn_err(CE_WARN, "In IOCTL ABORT READ.");
			asp->abort = 1;
			break;

		case ASTROPCI_FLUSH_REPLY_BUFFER:
		    {
			uint32_t reply_status = 0;
		    	int reply_value = 0;
			int x;
			
			/* Flush the reply buffer FIFO on the PCI DSP. 3 is the number of words the FIFO can hold */
   			for (x=0; x<3; x++) {
	  			reply_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hstr);
   				if (reply_status & REPLY_DATA_EXISTS)
  					reply_value = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->reply_buffer);
  				else
  					break;
  			}
			
			/* Clear the driver's internal reply variable */
			asp->reply = REPLY_BUFFER_EMPTY;
		     }
		     break;
			      
    		case ASTROPCI_GET_CONFIG_INFO:
        	    {
       			 uint16_t value[6];
        
			#if 0
			/* ddi_getw(handle0, (ushort_t *)configp + CONFIG_VENDOR_ID_OFFSET) */
        		value[0] = pci_config_getw(asp->config_access_handle, 0x00);	/* vendor_id */
        		value[1] = pci_config_getw(asp->config_access_handle, 0x02);	/* device_id */
        		value[2] = pci_config_getw(asp->config_access_handle, 0x04);	/* command_reg */
        		value[3] = pci_config_getw(asp->config_access_handle, 0x06);	/* status_reg */
        		value[4] = pci_config_getw(asp->config_access_handle, 0x10);	/* base address 0 low */
        		value[5] = pci_config_getw(asp->config_access_handle, 0x12);	/* base address 0 high */
        		#endif
        		
        		if (ddi_copyout((caddr_t) &value, (caddr_t) arg, sizeof(value), mode) != 0)
           			retval = EFAULT;
     		    }
     		    break;

     		default:
      			retval = ENOTTY;
      			break;
	}

 	mutex_exit (&asp->mu);

  	return retval;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_INTR
 
 PURPOSE:  Entry point. Interrupt handler.
 
 RETURNS:  Returns DDI_INTR_CLAIMED if interrupt belongs to device, else it
 	   returns DDI_INTR_UNCLAIMED.
 
 NOTES:    intr_status before reply buffer check is commented out to test for
 	   unclaimed interrupt problem.
*******************************************************************************/
static u_int astropci_intr (caddr_t instance)
{
	register struct image_buf *ip;
  	register struct status_buf *sp;
  	uint32_t intr_status = 0;
  	uint32_t transfer_status = 0;
  
  	/* Get current soft state */
  	struct astropci_state *asp = ddi_get_soft_state(statep, (int)instance);

  	if (asp == NULL)
  		return DDI_INTR_UNCLAIMED;

  	/* Allow only one thread at a time */
  	mutex_enter (&asp->mu);
  	while (asp->intr_busy) {
		cv_wait (&asp->intrcv, &asp->mu);
  	}
  	asp->intr_busy = 1;
  	mutex_exit (&asp->mu);

  	mutex_tryenter (&asp->mu);

  	/* Get the status of the current interrupt */
  	intr_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hstr);

  	/* Check that it is our device that is interrupting */
  	if (!(intr_status & DMA_INTERRUPTING)) {
  		mutex_exit(&asp->mu);
  	  	asp->intr_busy = 0;
		cv_signal(&asp->intrcv);
		return DDI_INTR_UNCLAIMED;
	}
	
	/* PCI board is generating a stray interrupt */
	else if ((intr_status & DMA_INTERRUPTING) && !(intr_status & (REPLY_INTERRUPT | IMAGE_BUFFER1_INTERRUPT | IMAGE_BUFFER2_INTERRUPT))) {
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_IMAGE_BUFF1_INTERRUPT);
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_IMAGE_BUFF2_INTERRUPT);
		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_REPLY_INTERRUPT);
		mutex_exit(&asp->mu);
  		asp->intr_busy = 0;
		cv_signal(&asp->intrcv);
		return DDI_INTR_CLAIMED;
	}

	/* Check for a readout abort */
	if ((intr_status & ABORT_INTERRUPT) == ABORT_INTERRUPT) {
		/* Clear the interrupt */
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_REPLY_INTERRUPT);
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_IMAGE_BUFF1_INTERRUPT);
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_IMAGE_BUFF2_INTERRUPT);

		/* Clear the image transfer status */
		transfer_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hctr);
		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr, (uint32_t)(transfer_status & IMAGE_BUFF1_CLEAR & IMAGE_BUFF2_CLEAR));

  		/* Set the abort flag */
  		asp->abort = 1;
  		cv_signal(&asp->bufcv);
		mutex_exit(&asp->mu);
  		asp->intr_busy = 0;
		cv_signal(&asp->intrcv);
		return DDI_INTR_CLAIMED;
  	}

  	/* Check if this is a reply interrupt and handle accordingly */
  	if ((intr_status & REPLY_INTERRUPT) && (intr_status & REPLY_DATA_EXISTS)) {
   		/* Clear the interrupt */
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_REPLY_INTERRUPT);
	 	asp->reply = ddi_get32(asp->reg_access_handle, (uint32_t *)&asp->regp->reply_buffer);
  		cv_signal(&asp->reply_bufcv);
 	}

  	/* Check if the PCI board is in some weird state and generated a reply interrupt without data. */
  	else if ((intr_status & REPLY_INTERRUPT) && !(intr_status & REPLY_DATA_EXISTS)) {
   		/* Clear the interrupt */
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_REPLY_INTERRUPT);
 	}

  	/* Otherwise, it may be an image transfer interrupt */
  	/* Check that image buffer1 is interrupting. */
  	if (intr_status & IMAGE_BUFFER1_INTERRUPT) {
		/* Set the busy flag */
  		transfer_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hctr);
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr, (uint32_t)(transfer_status | IMAGE_BUFF1_BUSY));
	
		/* Clear the interrupt */
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_IMAGE_BUFF1_INTERRUPT);

		/* Set DMA flags and signal a full buffer */
		asp->dma_counter++;
		asp->dma_done = 1;
		cv_signal(&asp->bufcv);
	}

  	/* Check that image buffer2 is interrupting. */
  	if (intr_status & IMAGE_BUFFER2_INTERRUPT) {
		/* Set the busy flag */
  		transfer_status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hctr);
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hctr, (uint32_t)(transfer_status | IMAGE_BUFF2_BUSY));
	
		/* Clear the interrupt */
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->hcvr, (uint32_t)CLEAR_IMAGE_BUFF2_INTERRUPT);

		/* Set DMA flags and signal a full buffer */
		asp->dma_counter++;
		asp->dma_done = 1;
		cv_signal(&asp->bufcv);
	}

  	mutex_exit (&asp->mu);
  	asp->intr_busy = 0;
	cv_signal(&asp->intrcv);

  	return DDI_INTR_CLAIMED;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_CHECK_DSP_BUFFER
 
 PURPOSE:  Check that the DSP input buffer (fifo) is not full.
 
 RETURNS:  Returns 1 if bit 1 is set and buffer is available for input.
 	   Returns 0 if bit 1 is unset and buffer is unavailable for input.
 
 NOTES:    This function must be called before writing to any resister in the
 	   astropci_regs structure. Otherwise, data may be overwritten in the
 	   DSP input buffer, since the DSP cannot keep up with the input rate.
 	   This function will exit after "max_tries". This will help prevent
 	   the system from hanging in case the PCI DSP hangs.
*******************************************************************************/
static int astropci_check_dsp_buffer(struct astropci_state *asp)
{
	uint32_t buffer_status = 0;
	uint32_t status = 0;
	const uint32_t buffer_ok_mask = BUFFER_OK_MASK;
   	const int max_tries = 100;
   	int try = 0;
	
	/* Get the status of the DSP input buffer */
	status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hstr);
	buffer_status = (status & buffer_ok_mask);
	
	/* Wait until buffer is ready to receive data */
	while (buffer_status != 2) {	/* WAS A 1 */
		if (try > max_tries)
			break;

		status = ddi_get32(asp->reg_access_handle,(uint32_t*)&asp->regp->hstr);
		buffer_status = (status & buffer_ok_mask);
		try++;
	}

	return buffer_status;
}

/*******************************************************************************
 FUNCTION: ASTROPCI_SET_BUFFER_ADDRESSES
 
 PURPOSE:  Set the image buffer addresses. This function allocates two memory
 	   regions and passes the their addresses to the DSP. These buffers are
 	   used by the "read()" entry point to temporarily store the image data.
 	   The buffers are "filled" by the PCI DSP.
 
 RETURNS:  Returns 0 upon success. Returns ENOMEM if the DMA memory allocation
 	   fails.
 
 NOTES:    This function must be called at the before taking any exposures.
*******************************************************************************/
static int astropci_set_buffer_addresses(struct astropci_state *asp)
{
	struct image_buf *ip;
	uint_t cookies;
	int value = 0;
	uint32_t reply = 0;
	int j = 0;

  	/* Initialize the image buffer dma handles and states */
  	for (j=0; j<2; j++) {
		asp->image_buf[j].buf_handle = (caddr_t) 0;
		asp->image_buf[j].state = image_buf_waiting;
  	}
  
  	/* Allocate the image buffer memory and handles for the dma transfer */ 
  	for (j=0; j<2; j++) {
		ip = &asp->image_buf[j];
		if ( ddi_dma_alloc_handle(asp->dip, &dma_attrs, DDI_DMA_DONTWAIT, NULL, &ip->buf_handle) != DDI_SUCCESS) {
			cmn_err (CE_WARN,"astropci_read: ddi_dma_alloc_handle failed");
			value = ENOMEM;
			goto EXITBUFADDR;
  		}

  		if ( ddi_dma_mem_alloc ( ip->buf_handle, image_buf_size, &access_attr, DDI_DMA_STREAMING, DDI_DMA_DONTWAIT, NULL, &ip->address, &ip->length, &ip->data_access_handle) != DDI_SUCCESS) {
			cmn_err (CE_WARN,"astropci_read: ddi_dma_mem_alloc failed");
			value = ENOMEM;
			goto EXITBUFADDR;
  		}
	
  		if ( ddi_dma_addr_bind_handle ( ip->buf_handle, NULL, ip->address, ip->length, DDI_DMA_READ, DDI_DMA_DONTWAIT, NULL, &ip->buf_cookie, &cookies) != DDI_DMA_MAPPED) {
			cmn_err (CE_WARN,"astropci_read: ddi_dma_addr_bind_handle failed");
			value = ENOMEM;
			goto EXITBUFADDR;
  		}
  	}

  	/* Save the DMA kernel buffer addresses. */
  	ip = &asp->image_buf[0];
  	asp->image_buffer_address1 = (uint32_t)ip->buf_cookie.dmac_address;
  
  	ip = &asp->image_buf[1];
  	asp->image_buffer_address2 = (uint32_t)ip->buf_cookie.dmac_address;
	
	/* Pass the DMA kernel buffer addresses to the DSP */
  	if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->dma_addr1, (uint32_t)asp->image_buffer_address1);

  	if (astropci_check_dsp_buffer(asp) == BUFFER_OK_MASK)
  		ddi_put32(asp->reg_access_handle, (uint32_t *)&asp->regp->dma_addr2, (uint32_t)asp->image_buffer_address2);

	#if DEBUG
	cmn_err (CE_WARN, "buffer address1 = 0x%08X", asp->image_buffer_address1);
	cmn_err (CE_WARN, "buffer address2 = 0x%08X", asp->image_buffer_address2);
	#endif
	
EXITBUFADDR:
	if (value == ENOMEM) {
  		for (j=0; j<2; j++) {
  			if (asp->image_buf[j].buf_handle != (caddr_t) -1) {
     				ip = &asp->image_buf[j];
     				if ( ddi_dma_unbind_handle(ip->buf_handle) != DDI_SUCCESS )
         				cmn_err(CE_WARN,"ddi_dma_unbind_handle failed");
     				else {
         				ddi_dma_mem_free(&ip->data_access_handle);
         				ddi_dma_free_handle(&ip->buf_handle);
     				}
  			}
  		}
  	}

	return value;
}
