/*******************************************************************************
*   File:	astropci_impl.h
*   Author:   	Scott Streit
*   Abstract: 	Private implementation details for Solaris device driver for
*             	the SDSU PCI Interface Card.
*
*
*   Revision History:     Date      Who   Version    Description
*   --------------------------------------------------------------------------
*                       11/12/99    sds     1.0      Initial Release
*
*******************************************************************************/
#ifndef __ASTROPCI_IMPL_H__
#define __ASTROPCI_IMPL_H__

#include <sys/types.h>

/*******************************************************************************
 	Define the ID string stored in the FCode PROM
*******************************************************************************/
const char prom_id_string_one[] = "pci1057,1801";
const char prom_id_string_two[] = "astropci";

/*******************************************************************************
	DSP/PCI Control Registers
*******************************************************************************/
struct astropci_regs {
	volatile uint32_t reserved1;	/* DSP Reserved register		*/
	volatile uint32_t reserved2;	/* DSP Reserved register		*/
	volatile uint32_t reserved3;	/* DSP Reserved register		*/
	volatile uint32_t reserved4;	/* DSP Reserved register		*/
	volatile uint32_t hctr;		/* 0x10 Host interface control register on DSP */
  	volatile uint32_t hstr;		/* 0x14 Host interface status register on DSP */
  	volatile uint32_t hcvr;		/* 0x18 Host command vector register on DSP */
  	volatile uint32_t reply_buffer;	/* 0x1C Command reply buffer            */
  	volatile uint32_t manual_cmd;	/* 0x20 DSP command register 	        */
 	volatile uint32_t board_dest;	/* 0x24 Board destination (PCI, Timing, Utility)*/
  	volatile uint32_t arg1;		/* 0x28 DSP command argument #1 	*/
  	volatile uint32_t arg2;		/* 0x2C DSP command argument #2 	*/
 	volatile uint32_t arg3;		/* 0x30 DSP command argument #3 	*/
  	volatile uint32_t arg4;		/* 0x34 DSP command argument #4 	*/
  	volatile uint32_t arg5;		/* 0x38 DSP command argument #5 	*/
  	volatile uint32_t unused;	/* 0x3C Unused	 	*/
  	volatile uint32_t dma_addr1;  	/* 0x40 DMA kernel buffer 1 address	*/
 	volatile uint32_t dma_addr2;	/* 0x44 DMA kernel buffer 2 address	*/
};

/*******************************************************************************
	Camera Control Table
*******************************************************************************/
struct astropci_camregs {
  	volatile uint32_t exptime;	   /* 0x80 Image exposure time	        */
  	volatile uint32_t util_options;    /* 0x84 Utility board options        */
  				   	   /*      1 = open shutter (image)     */
  				   	   /*      0 = close shutter (dark)     */
  	volatile uint32_t cam_status;	   /* 0x88 Camera status		*/
  	volatile uint32_t ncols;	   /* 0x8C Number of columns on the CCD */
  	volatile uint32_t nrows;	   /* 0x90 Number of rows on the CCD    */
  	volatile uint32_t ncols_bin;	   /* 0x94 Number of columns binned     */
  	volatile uint32_t nrows_bin;	   /* 0x98 Number of rows binned        */
};

/*******************************************************************************
	ddi_device_acc_attr - Data access attributes structure. Describes the
			      data  access characteristics and requirements of
			      the device.
*******************************************************************************/
static ddi_device_acc_attr_t access_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};

/*******************************************************************************
	ddi_dma_attr - DMA attributes structure. Describes device and  DMA  engine
     		       specific  attributes necessary to allocate DMA resources for
		       a device.
*******************************************************************************/
static ddi_dma_attr_t dma_attrs = {
        DMA_ATTR_V0,                    /* Version number, Solaris default */
        (uint64_t)0x00000000, 		/* Low DMA address range           */
        (uint64_t)0xffffffff, 		/* High DMA address range 	   */
        (uint64_t)0x00ffffff,   	/* DMA byte counter max 	   */
        (uint64_t)0x1, 		        /* DMA address alignment	   */
        (uint_t)0x3f,               	/* Burst size of 1,2,4,16,32 (0x3F [63 dec] = sum of burst sizes [1+2+4+8+16+32]) */
        (uint32_t)0x1,                  /* Minimum transfer size 	   */
        (uint64_t)0x1fffffe0, 		/* Maximum transfer size 512Mb 	   */
        (uint64_t)0x00ffffff,   	/* Segment boundary		   */
        (int)50,          		/* Scatter/gather list lgth 	   */ 
        (uint32_t)512,  		/* Granularity 16b		   */
        (uint_t)0                       /* Flags, Solaris default 	   */
};


/*******************************************************************************
	image_buf - Structure that describes the kernel image buffers. Used to
		    temporarily store the image data before transfer to user
		    space.
*******************************************************************************/
struct image_buf {
  caddr_t		address;		/* Image buffer address in kernel memory for the DMA transfer */
  int			operation;		/* Current DMA operation, DDI_DMA_READ or DDI_DMA_WRITE */
  long			size;			/* The (maximum) number of bytes to be transfered to the image buffer */
  ulong_t  		length;			/* The actual number of bytes of kernel memory allocated for the image buffer */
  u_int			state;			/* Current state of DMA transfer, image_buf_waiting, image_buf_ready, image_buf_done, image_buf_error, image_buf_busy */
  ddi_dma_cookie_t	buf_cookie;		/* Structure that holds DMA info for this image buffer, dmac_address, dmac_size */
  ddi_dma_handle_t	buf_handle;		/* Handle to the allocated DMA handle for the image buffer */
  ddi_acc_handle_t 	data_access_handle;	/* Handle to the allocated kernel memory for the DMA transfer */
};

/*******************************************************************************
	status_buf - Structure that describes the kernel status buffer. Used to
		     determine the current status of the DSP.
*******************************************************************************/
struct status_buf {
  caddr_t		status_address;		/* Status buffer address in kernel memory for the DMA transfer */
  long			status_size;		/* The (maximum) number of bytes to be transfered to the status buffer */
  ulong_t			status_length;		/* The actual number of bytes of kernel memory allocated for the status buffer */
  ddi_dma_cookie_t	status_buf_cookie;	/* Structure that holds DMA info for this status buffer, dmac_address, dmac_size */
  ddi_dma_handle_t	status_buf_handle;	/* Handle to the allocated DMA handle for the status buffer */
  ddi_acc_handle_t 	status_access_handle;	/* Handle to the allocated kernel memory for the DMA transfer */
};

/*******************************************************************************
	astropci_state - Driver state structure.  All state variables related
			 to the state of the driver are kept in here.
*******************************************************************************/
struct astropci_state {
	dev_info_t *dip;		/* pointer to devices dev_info node */

  	struct astropci_regs *regp;		/* DSP/PCI control registers */
  	struct astropci_camregs *camregp;	/* Camera control table      */
 
  	volatile int open;		/* Flag set when driver opened              */
  	volatile int busy;		/* Set when transfers in progress           */
  	volatile int intr_busy;		/* Set when interrupt service in progress   */

  	/* interrupt control `cookie's */
  	ddi_iblock_cookie_t  iblock_cookie;
  	ddi_idevice_cookie_t idevice_cookie;

  	/* dma buffers */
  	struct buf *bp;
  	ddi_dma_handle_t handle;
  	struct image_buf  image_buf[3];
  	int dma_done;
  	int dma_counter;
  	int buf_numbusy;
  	int num_of_bufs_in_use;
  	int buf_done;
  	int buf_ready;
  	int buf_busy;
  	int buf_waiting;
	int abort;			/* Specifies whether or not to abort the read. 1=abort, 0=normal */

  	/* Reply buffer */
	int reply;
  	
  	/* Status buffer */
  	struct status_buf status_buffer;

  	/* Image readout progress buffer */
  	int progress_buffer;
  
  	/* thread locking and synchronization */
  	kmutex_t mu;		  /* Mutual exclusion lock */
  	kmutex_t reply_mu;	  /* Mutual exclusion lock for reply buffer */
  	kcondvar_t cv;	  	  /* Synchronization condition variable */
  	kcondvar_t reply_cv;	  /* Reply buffer synchronization condition variable */
  	kcondvar_t bufcv;	  /* Image_buf synchronization condition variable */
  	kcondvar_t reply_bufcv;   /* Reply buffer synchronization condition variable */
  	kcondvar_t intrcv;	  /* ISR synchronization condition variable */

  	long byteremain;				/* Size of transfer remaining to be transfered */
  	ddi_acc_handle_t config_access_handle;		/* Config registers access handle         */
  	ddi_acc_handle_t reg_access_handle;		/* DSP/PCI control register access handle */
  	ddi_acc_handle_t camreg_access_handle;		/* Camera control register access handle  */

  	/* Image Buffer Addresses */
  	uint32_t image_buffer_address1;
  	uint32_t image_buffer_address2;
};

/* Time out a transfer if no dma interrupt occurrs within this time.  */
/* Must be at least long enough to transfer minphys bytes.  The timer */
/* is restarted for each new block during a large transfer.           */
const int dma_read_timeout_usec = 60000000; 	/* 60.0 secs   */
const int read_reply_timeout_usec = 10000000; 	/* 10.0 secs   */
const uint_t image_buf_size = 131080;  		/* DSP Fixed   */
const int image_buf_busy = 3;
const int image_buf_waiting = 4;

#endif

